use crate::{error::Error, spell::Spell, state::State};
use derive_more::Display;

#[derive(Debug, Display, Clone, Copy, PartialEq, Eq, Hash)]
pub struct BlockId(u64);

pub struct Block {
    internal_id: u64,
    name: Option<String>,
    spells: Vec<Box<dyn Spell>>,
}

impl Default for Block {
    fn default() -> Self {
        use std::sync::atomic::{AtomicU64, Ordering};
        static INTERNAL_ID: AtomicU64 = AtomicU64::new(1);
        match INTERNAL_ID.fetch_add(1, Ordering::Relaxed) {
            u64::MAX => panic!("reached the max number of blocks, how did you did it?"),
            internal_id => Self {
                internal_id,
                name: Default::default(),
                spells: Default::default(),
            },
        }
    }
}

impl Block {
    pub fn with_name(self, name: impl Into<String>) -> Self {
        let name = name.into();
        Self {
            name: (!name.is_empty()).then_some(name),
            ..self
        }
    }

    pub fn with_spells(self, spells: impl IntoIterator<Item = Box<dyn Spell>>) -> Self {
        Self {
            spells: spells.into_iter().collect(),
            ..self
        }
    }

    pub fn into_parts(self) -> (BlockId, Self) {
        (self.id(), self)
    }

    pub fn id(&self) -> BlockId {
        BlockId(self.internal_id)
    }

    pub fn iter(&self) -> impl Iterator<Item = &dyn Spell> {
        self.spells.iter().map(Box::as_ref)
    }

    pub fn indexed_iter(&self) -> impl Iterator<Item = &dyn Spell> {
        self.spells.iter().map(Box::as_ref)
    }

    pub fn name(&self) -> Option<&str> {
        self.name.as_deref()
    }

    pub fn try_cast_spells(&self, state: State) -> Result<State, Error> {
        self.spells
            .iter()
            .try_fold(state, |state, spell| spell.cast(state))
    }
}
