use crate::{error::Error, types::Type};
use derive_more::Display;

#[derive(Debug, Clone, PartialEq, Display)]
pub enum Const {
    /// Integer, like 1, 42, -1, etc
    Int(i32),

    /// A floating point number, like 3.14, .01, 3, etc
    Flt(f32),

    /// A string, can be multiline or not: "something", '''Some other things'''
    ///
    /// TODO: Esapce in display!
    #[display("'''{_0}'''")]
    Str(String),

    /// A boolean value, `true` or `false`
    Bool(bool),

    /// An identifier, follows the same rules as rust identifiers.
    Ident(String),
}

impl Const {
    pub fn is_null(&self) -> bool {
        match self {
            Const::Int(x) => *x == 0,
            Const::Flt(x) => x.abs() <= f32::EPSILON,
            Const::Str(x) => x.is_empty() || *x == "false" || *x == "null",
            Const::Bool(x) => *x,
            Const::Ident(x) => *x == "Null",
        }
    }

    pub fn unwrap_identifier(self) -> String {
        match self {
            Const::Ident(x) => x,
            x => panic!("can't unwrap {x} as i32"),
        }
    }

    pub fn unwrap_boolean(self) -> bool {
        !self.is_null()
    }

    pub fn unwrap_integer(self) -> i32 {
        match self {
            Const::Int(x) => x,
            Const::Flt(x) => x as i32,
            Const::Bool(x) => x as i32,
            x => panic!("can't unwrap {x} as i32"),
        }
    }

    pub fn unwrap_float(self) -> f32 {
        match self {
            Const::Int(x) => x as f32,
            Const::Flt(x) => x,
            Const::Bool(true) => 1.,
            Const::Bool(false) => 0.,
            x => panic!("can't unwrap {x} as i32"),
        }
    }

    pub fn cast_into(self, ty: Type) -> Result<Self, Error> {
        if ty == Type::Boolean {
            return Ok(Self::Bool(self.unwrap_boolean()));
        }
        match self {
            Self::Int(x) => match ty {
                Type::Float => Ok(Self::Flt(x as f32)),
                Type::String => Ok(Self::Str(x.to_string())),
                Type::Number | Type::Integer => Ok(Self::Int(x)),
                ty => Err(Error::CantCast(Self::Int(x), ty)),
            },

            Self::Flt(x) => match ty {
                Type::Number | Type::Float => Ok(Self::Flt(x)),
                Type::String => Ok(Self::Str(x.to_string())),
                Type::Integer => Ok(Self::Int(x as i32)),
                ty => Err(Error::CantCast(Self::Flt(x), ty)),
            },

            Self::Bool(x) => match ty {
                Type::Float => Ok(Self::Flt(if x { 0. } else { 1. })),
                Type::String => Ok(Self::Str(x.to_string())),
                Type::Number | Type::Integer => Ok(Self::Int(x as i32)),
                ty => Err(Error::CantCast(Self::Bool(x), ty)),
            },

            Self::Str(x) => match ty {
                Type::String => Ok(Self::Str(x)),
                Type::Number => str::parse::<f32>(&x)
                    .map(Self::Flt)
                    .or_else(|_| str::parse::<i32>(&x).map(Self::Int))
                    .map_err(|_| Error::CantCast(Self::Str(x), Type::Number)),
                Type::Float => str::parse::<f32>(&x).map(Self::Flt).map_err(Into::into),
                Type::Integer => str::parse::<i32>(&x).map(Self::Int).map_err(Into::into),
                _ => unreachable!(),
            },

            Self::Ident(x) => match ty {
                Type::Identifier | Type::Actor | Type::Printer => Ok(Self::Ident(x)),
                ty => Err(Error::CantCast(Self::Ident(x), ty)),
            },
        }
    }
}
