use derive_more::Display;

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Display)]
#[display("{op}", op = self.as_str())]
pub enum Unary {
    Not,
    Neg,
}

impl Unary {
    pub fn as_str(&self) -> &'static str {
        match self {
            Unary::Not => "!",
            Unary::Neg => "-",
        }
    }
}

impl AsRef<str> for Unary {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Display)]
#[display("{op}", op = self.as_str())]
pub enum Numeric {
    Add,
    Sub,
    Mul,
    Div,
    Mod,
    CompGE,
    CompGT,
    CompLT,
    CompLE,
}

impl Numeric {
    pub fn as_str(&self) -> &'static str {
        match self {
            Numeric::Add => "+",
            Numeric::Sub => "-",
            Numeric::Mul => "*",
            Numeric::Div => "/",
            Numeric::Mod => "%",
            Numeric::CompGE => ">=",
            Numeric::CompGT => ">",
            Numeric::CompLT => "<=",
            Numeric::CompLE => "<",
        }
    }
}

impl AsRef<str> for Numeric {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Display)]
#[display("{op}", op = self.as_str())]
pub enum Logic {
    Or,
    And,
    Xor,
}

impl Logic {
    pub fn as_str(&self) -> &'static str {
        match self {
            Logic::Or => "|",
            Logic::And => "&",
            Logic::Xor => "^",
        }
    }
}

impl AsRef<str> for Logic {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Display)]
#[display("{op}", op = self.as_str())]
pub enum Equal {
    CompEQ,
    CompNEQ,
}

impl Equal {
    pub fn as_str(&self) -> &'static str {
        match self {
            Equal::CompEQ => "=",
            Equal::CompNEQ => "!=",
        }
    }

    pub fn apply(self, eq_res: bool) -> bool {
        (self.check_equals() && eq_res) || (self.check_not_equals() && !eq_res)
    }

    pub fn check_equals(self) -> bool {
        !self.check_not_equals()
    }

    pub fn check_not_equals(self) -> bool {
        matches!(self, Equal::CompNEQ)
    }
}

impl AsRef<str> for Equal {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Display)]
#[display("{op}", op = self.as_str())]
pub enum Binary {
    Numeric(Numeric),
    Logic(Logic),
    Equal(Equal),
}

impl Binary {
    pub fn as_str(&self) -> &'static str {
        match self {
            Binary::Numeric(op) => op.as_str(),
            Binary::Logic(op) => op.as_str(),
            Binary::Equal(op) => op.as_str(),
        }
    }
}

impl AsRef<str> for Binary {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl From<Numeric> for Binary {
    fn from(value: Numeric) -> Self {
        Binary::Numeric(value)
    }
}

impl From<Logic> for Binary {
    fn from(value: Logic) -> Self {
        Binary::Logic(value)
    }
}

impl From<Equal> for Binary {
    fn from(value: Equal) -> Self {
        Binary::Equal(value)
    }
}
