use crate::{
    ast::{Block, BlockId, Expression},
    error::Error,
};
use derive_more::Display;
use std::{
    collections::{HashMap, HashSet},
    hash::Hash,
};

#[derive(Debug, Display)]
#[display("Edge({_0} -> {_1})")]
pub struct Edge(BlockId, BlockId, Option<Expression>);

impl Edge {
    pub fn new(from: BlockId, to: BlockId) -> Self {
        Self(from, to, None)
    }

    pub fn with_condition(from: BlockId, to: BlockId, predicate: Expression) -> Self {
        Self(from, to, Some(predicate))
    }
}

impl Eq for Edge {}
impl PartialEq for Edge {
    fn eq(&self, other: &Self) -> bool {
        self.0 == other.0 && self.1 == other.1
    }
}

impl Hash for Edge {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.0.hash(state);
        self.1.hash(state);
    }
}

#[derive(Default)]
pub struct Graph {
    nodes: HashMap<BlockId, Block>,
    edges: HashSet<Edge>,
    id_mapping: HashMap<String, BlockId>,
}

impl Graph {
    /// Note, can't fail becose block's ids are unique!
    pub fn add_node(&mut self, block: Block) -> BlockId {
        let (id, block) = block.into_parts();

        if let Some(name) = block.name() {
            self.id_mapping.insert(name.to_string(), id);
        }
        self.nodes.insert(id, block);

        id
    }

    pub fn get_node(&self, id: BlockId) -> Result<&Block, Error> {
        self.nodes
            .get(&id)
            .ok_or(Error::Undefined(format!("block id {id}").into()))
    }

    pub fn get_node_by_name(&self, name: impl AsRef<str>) -> Result<&Block, Error> {
        self.get_node(*self.id_mapping.get(name.as_ref()).ok_or(Error::Undefined(
            format!("block with name {}", name.as_ref()).into(),
        ))?)
    }

    pub fn mut_node(&mut self, id: BlockId) -> Result<&mut Block, Error> {
        self.nodes
            .get_mut(&id)
            .ok_or(Error::Undefined(format!("block id {id}").into()))
    }

    pub fn mut_node_by_name(&mut self, name: impl AsRef<str>) -> Result<&mut Block, Error> {
        self.mut_node(*self.id_mapping.get(name.as_ref()).ok_or(Error::Undefined(
            format!("block with name {}", name.as_ref()).into(),
        ))?)
    }

    pub fn add_edge(&mut self, from: BlockId, to: BlockId) -> Result<(), Error> {
        self.edges
            .insert(Edge::new(from, to))
            .then_some(())
            .ok_or_else(|| Error::Redefined(format!("edge from block {from} to block {to}").into()))
    }

    pub fn add_conditional_edge(
        &mut self,
        from: BlockId,
        to: BlockId,
        predicate: Expression,
    ) -> Result<(), Error> {
        self.edges
            .insert(Edge::with_condition(from, to, predicate))
            .then_some(())
            .ok_or_else(|| Error::Redefined(format!("edge from block {from} to block {to}").into()))
    }

    pub fn successors(&self, from: BlockId) -> Vec<(BlockId, Option<&Expression>)> {
        (self.edges.iter())
            .filter_map(|Edge(f, to, pred)| (*f == from).then_some((*to, pred.as_ref())))
            .collect()
    }

    pub fn conditional_successors(&self, from: BlockId) -> Vec<(BlockId, &Expression)> {
        (self.edges.iter())
            .flat_map(|Edge(f, to, pred)| {
                (*f == from)
                    .then(|| pred.as_ref().map(|pred| (*to, pred)))
                    .flatten()
            })
            .collect()
    }

    pub fn unconditional_successor(&self, from: BlockId) -> Result<Option<BlockId>, Error> {
        if self.edges.iter().filter(|Edge(f, ..)| *f == from).count() <= 1 {
            return Err(Error::Unexpected(
                "got multiple unconditional successors for a single block".into(),
            ));
        }

        Ok((self.edges.iter()).find_map(|Edge(f, to, _)| (*f == from).then_some(*to)))
    }
}
