use crate::{
    ast::{AstSpan, Expression, VarOrConst, operator},
    parser::*,
};
use nom::{
    Parser,
    branch::alt,
    bytes::complete::tag,
    character::complete::{char, multispace0},
    combinator::{map, value},
    multi::many0,
    sequence::preceded,
};

fn var_or_constant(s: Span) -> ParserResult<Expression> {
    // Note: the order is important!
    utils::map_with_locaiton(
        alt((
            map(utils::integer, VarOrConst::integer),
            map(utils::float, VarOrConst::float),
            map(utils::string, VarOrConst::string),
            value(VarOrConst::boolean(true), utils::keyword(keywords::TRUE)),
            value(VarOrConst::boolean(false), utils::keyword(keywords::FALSE)),
            map(preceded(char('$'), utils::identifier), VarOrConst::Var),
            map(utils::identifier, VarOrConst::identifier),
        )),
        Expression::Leaf,
    )
    .parse(s)
}

fn leaf(s: Span) -> ParserResult<Expression> {
    // Note: the order is important!
    alt((
        unop(operator::Unary::Not, expression),
        unop(operator::Unary::Neg, expression),
        utils::parse_paren(expression),
        var_or_constant,
    ))
    .parse(s)
}

fn unop<'a>(
    op: operator::Unary,
    next: impl Parser<Span<'a>, Output = Expression, Error = ParserError>,
) -> impl Parser<Span<'a>, Output = Expression, Error = ParserError> {
    utils::map_with_locaiton(
        (multispace0, tag(op.as_str()), next),
        move |ast_span, (.., expr)| Expression::Unary(ast_span, op, Box::new(expr)),
    )
}

fn binop<'a>(
    op: impl Into<operator::Binary>,
    next: impl Parser<Span<'a>, Output = Expression, Error = ParserError>,
) -> impl Parser<Span<'a>, Output = (AstSpan, operator::Binary, Expression), Error = ParserError> {
    // Used to mitigate the impact of monomorphisation.
    fn aux<'a>(
        op: operator::Binary,
        next: impl Parser<Span<'a>, Output = Expression, Error = ParserError>,
    ) -> impl Parser<Span<'a>, Output = (AstSpan, operator::Binary, Expression), Error = ParserError>
    {
        preceded(
            multispace0,
            utils::with_location((map(tag(op.as_str()), move |_| op), next))
                .map(|(ast_span, (op, expr))| (ast_span, op, expr)),
        )
    }

    aux(op.into(), next)
}

fn fold_exprs(
    initial: Expression,
    remainder: Vec<(AstSpan, operator::Binary, Expression)>,
) -> Expression {
    remainder
        .into_iter()
        .fold(initial, |acc, (ast_span, op, expr)| {
            Expression::Binary(ast_span, Box::new(acc), op, Box::new(expr))
        })
}

fn terms(s: Span) -> ParserResult<Expression> {
    let (s, initial) = leaf(s)?;
    let (s, remainder): (Span, Vec<_>) = many0(alt((
        binop(operator::Numeric::Mul, leaf),
        binop(operator::Numeric::Div, leaf),
        binop(operator::Numeric::Mod, leaf),
    )))
    .parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}

fn sums(s: Span) -> ParserResult<Expression> {
    let (s, initial) = terms(s)?;
    let (s, remainder) = many0(alt((
        binop(operator::Numeric::Add, terms),
        binop(operator::Numeric::Sub, terms),
    )))
    .parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}

fn logic_terms(s: Span) -> ParserResult<Expression> {
    let (s, initial) = sums(s)?;
    let (s, remainder) = many0(binop(operator::Logic::And, sums)).parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}

fn logic_sums(s: Span) -> ParserResult<Expression> {
    let (s, initial) = logic_terms(s)?;
    let (s, remainder) = many0(alt((
        binop(operator::Logic::Xor, logic_terms),
        binop(operator::Logic::Or, logic_terms),
    )))
    .parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}

fn logic_comp(s: Span) -> ParserResult<Expression> {
    let (s, initial) = logic_sums(s)?;
    let (s, remainder) = many0(alt((
        binop(operator::Numeric::CompLE, logic_sums),
        binop(operator::Numeric::CompGE, logic_sums),
        binop(operator::Numeric::CompLT, logic_sums),
        binop(operator::Numeric::CompGT, logic_sums),
    )))
    .parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}

pub fn expression(s: Span) -> ParserResult<Expression> {
    let (s, initial) = logic_comp(s)?;
    let (s, remainder) = many0(alt((
        binop(operator::Equal::CompEQ, logic_comp),
        binop(operator::Equal::CompNEQ, logic_comp),
    )))
    .parse(s)?;
    Ok((s, fold_exprs(initial, remainder)))
}
