use crate::parser::Location;
use nom::{Compare, FindSubstring, Input, Offset};

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub struct Span<'a> {
    /// The offset represents the position of the fragment relatively to
    /// the input of the parser. It starts at offset 0.
    offset: usize,

    /// The line number of the fragment relatively to the input of the
    /// parser. It starts at line 1.
    line: usize,

    /// The fragment that is spanned.
    /// The fragment represents a part of the input of the parser.
    fragment: &'a str,
}

impl<'a> core::ops::Deref for Span<'a> {
    type Target = &'a str;
    fn deref(&self) -> &Self::Target {
        &self.fragment
    }
}

impl From<Span<'_>> for Location {
    fn from(value: Span<'_>) -> Self {
        value.into_location()
    }
}

impl From<Span<'_>> for String {
    fn from(value: Span<'_>) -> Self {
        value.into_fragment().to_string()
    }
}

impl<'a> From<Span<'a>> for &'a str {
    fn from(value: Span<'a>) -> Self {
        value.into_fragment()
    }
}

impl<'a> Span<'a> {
    pub fn new(fragment: &'a str) -> Span<'a> {
        Self {
            offset: 0,
            line: 1,
            fragment,
        }
    }

    pub fn location(&self) -> Location {
        let Self { offset, line, .. } = *self;
        Location { offset, line }
    }

    pub fn into_location(self) -> Location {
        let Self { offset, line, .. } = self;
        Location { offset, line }
    }

    /// The offset represents the position of the fragment relatively to the input of the parser.
    /// It starts at offset 0.
    pub fn location_offset(&self) -> usize {
        self.offset
    }

    /// The line number of the fragment relatively to the input of the parser. It starts at line 1.
    pub fn location_line(&self) -> usize {
        self.line
    }

    /// The fragment that is spanned. The fragment represents a part of the input of the parser.
    pub fn fragment(&self) -> &str {
        self.fragment
    }

    /// The fragment that is spanned. The fragment represents a part of the input of the parser.
    pub fn into_fragment(self) -> &'a str {
        self.fragment
    }

    fn slice_by(&self, next_fragment: &'a str) -> Self {
        match self.fragment.offset(next_fragment) {
            0 => Self {
                line: self.line,
                offset: self.offset,
                fragment: next_fragment,
            },
            consumed_len => Self {
                line: self.line
                    + memchr::Memchr::new(b'\n', self.fragment.take(consumed_len).as_bytes())
                        .count(),
                offset: self.offset + consumed_len,
                fragment: next_fragment,
            },
        }
    }
}

impl Offset for Span<'_> {
    fn offset(&self, second: &Self) -> usize {
        second.fragment().as_ptr() as usize - self.fragment().as_ptr() as usize
    }
}

impl<'a> Input for Span<'a> {
    type Item = <&'a str as Input>::Item;
    type Iter = <&'a str as Input>::Iter;
    type IterIndices = <&'a str as Input>::IterIndices;

    fn input_len(&self) -> usize {
        self.fragment.input_len()
    }

    fn take(&self, index: usize) -> Self {
        self.slice_by(self.fragment.take(index))
    }

    fn take_from(&self, index: usize) -> Self {
        self.slice_by(self.fragment.take_from(index))
    }

    fn take_split(&self, index: usize) -> (Self, Self) {
        (self.take_from(index), self.take(index))
    }

    fn position<P>(&self, predicate: P) -> Option<usize>
    where
        P: Fn(Self::Item) -> bool,
    {
        self.fragment.position(predicate)
    }

    fn iter_elements(&self) -> Self::Iter {
        self.fragment.iter_elements()
    }

    fn iter_indices(&self) -> Self::IterIndices {
        self.fragment.iter_indices()
    }

    fn slice_index(&self, count: usize) -> Result<usize, nom::Needed> {
        self.fragment.slice_index(count)
    }
}

impl Compare<&str> for Span<'_> {
    fn compare(&self, t: &str) -> nom::CompareResult {
        self.fragment().compare(t)
    }

    fn compare_no_case(&self, t: &str) -> nom::CompareResult {
        self.fragment().compare_no_case(t)
    }
}

impl FindSubstring<&str> for Span<'_> {
    fn find_substring(&self, substr: &str) -> Option<usize> {
        self.fragment().find_substring(substr)
    }
}
