use crate::{
    ast::AstSpan,
    parser::{error::*, keywords, span::*},
};
use nom::{
    Parser,
    branch::alt,
    bytes::{
        complete::{tag, take_while_m_n},
        take_until,
    },
    character::complete::{alpha1, alphanumeric1, char, digit1, multispace0, multispace1},
    combinator::{opt, peek, recognize, value, verify},
    multi::{fold, many0},
    sequence::{delimited, pair, preceded},
};

pub fn parse_paren<'a, T>(
    inner: impl Parser<Span<'a>, Output = T, Error = ParserError>,
) -> impl Parser<Span<'a>, Output = T, Error = ParserError> {
    preceded(multispace0, delimited(char('('), inner, char(')')))
}

pub fn with_location<'a, O>(
    mut parser: impl Parser<Span<'a>, Output = O, Error = ParserError>,
) -> impl Parser<Span<'a>, Output = (AstSpan, O), Error = ParserError> {
    move |span: Span<'a>| {
        let begin_s = multispace0(span)?.0;
        let (end_s, res) = parser.parse(span)?;
        Ok((end_s, (AstSpan::from((begin_s, end_s)), res)))
    }
}

pub fn map_with_locaiton<'a, F, O>(
    mut parser: F,
    mut cb: impl FnMut(AstSpan, <F as Parser<Span<'a>>>::Output) -> O,
) -> impl Parser<Span<'a>, Output = O, Error = ParserError>
where
    F: Parser<Span<'a>, Error = ParserError>,
{
    move |span: Span<'a>| {
        let begin_s = multispace0(span)?.0;
        let (end_s, res) = parser.parse(span)?;
        Ok((end_s, cb(AstSpan::from((begin_s, end_s)), res)))
    }
}

pub fn keyword<'a>(
    ident: &'static str,
) -> impl Parser<Span<'a>, Output = Span<'a>, Error = ParserError> {
    delimited(multispace1, tag(ident), multispace1)
}

pub fn float<'a>(s: Span<'a>) -> ParserResult<'a, f32> {
    let s = multispace0(s)?.0;
    let parse_f32 = |s: Span<'a>| str::parse::<f32>(s.into());
    pair(
        digit1.map_res(parse_f32),
        opt(recognize((char('.'), digit1)).map_res(parse_f32)),
    )
    .map(|(intp, fltp)| intp + fltp.unwrap_or_default())
    .parse(s)
}

pub fn identifier(s: Span) -> ParserResult<String> {
    let s = multispace0(s)?.0;
    recognize(pair(
        alt((alpha1, tag("_"))),
        many0(alt((alphanumeric1, tag("_")))),
    ))
    .map_res(|ident: Span| {
        keywords::check(ident.into_fragment())
            .map_err(|reserved| ParserError::UseOfReservedKeyword(ident.into(), reserved))
            .map(String::from)
    })
    .parse(s)
}

pub fn integer<'a>(s: Span<'a>) -> ParserResult<'a, i32> {
    let s = multispace0(s)?.0;
    digit1
        .map_res(|s: Span<'a>| str::parse::<i32>(s.into()))
        .parse(s)
}

pub fn string(span: Span) -> ParserResult<String> {
    #[derive(Clone, Copy)]
    enum StringFragment<'a> {
        Literal(&'a str),
        EscapedChar(char),
        EscapedWhiteSpaces,
    }

    #[derive(Clone, Copy)]
    enum StringDelimiterQuote {
        Single,
        Double,
        TripleSingle,
        TripleDouble,
    }

    impl StringDelimiterQuote {
        fn as_str(&self) -> &'static str {
            match self {
                StringDelimiterQuote::Double => "\"",
                StringDelimiterQuote::Single => "'",
                StringDelimiterQuote::TripleSingle => "'''",
                StringDelimiterQuote::TripleDouble => "\"\"\"",
            }
        }

        fn get_parser<'a>(self) -> impl Parser<Span<'a>, Output = Self, Error = ParserError> {
            value(self, tag(self.as_str()))
        }
    }

    let span = multispace0(span)?.0;
    let delimiter = peek(alt((
        StringDelimiterQuote::Double.get_parser(),
        StringDelimiterQuote::Single.get_parser(),
        StringDelimiterQuote::TripleSingle.get_parser(),
        StringDelimiterQuote::TripleDouble.get_parser(),
    )))
    .parse(span)?
    .1;

    // Parse an escaped character: \n, \t, \r, \u{00AC}, etc.
    let parse_unicode = delimited(
        tag("u{"),
        take_while_m_n(1, 6, |c: char| c.is_ascii_hexdigit()),
        char('}'),
    )
    .map(Span::into_fragment)
    .map_res(move |hex: &str| u32::from_str_radix(hex, 16))
    .map_opt(std::char::from_u32);

    let parse_escaped_char = preceded(
        char('\\'),
        alt((
            parse_unicode,
            // The `value` parser returns a fixed value (the first argument) if it parser (the
            // second argument) succeeds.
            value('\n', char('n')),
            value('\r', char('r')),
            value('\t', char('t')),
            value('\u{08}', char('b')),
            value('\u{0C}', char('f')),
            value('\\', char('\\')),
            value('/', char('/')),
            value('"', char('"')),
            value('\'', char('\'')),
        )),
    );

    let parse_fragment = alt((
        // Take until the end of the string
        verify(take_until(delimiter.as_str()), |s: &Span| !s.is_empty())
            .map(Span::into_fragment)
            .map(StringFragment::Literal),
        // Be aware of escaped characters
        parse_escaped_char.map(StringFragment::EscapedChar),
        // We handle escaped whitespaces
        value(
            StringFragment::EscapedWhiteSpaces,
            preceded(char('\\'), multispace1),
        ),
    ));

    let from_fragments = |mut string: String, fragment| {
        match fragment {
            StringFragment::Literal(s) => string.push_str(s),
            StringFragment::EscapedChar(c) => string.push(c),
            StringFragment::EscapedWhiteSpaces => {}
        }
        string
    };

    delimited(
        tag(delimiter.as_str()),
        fold(0.., parse_fragment, String::default, from_fragments),
        tag(delimiter.as_str()),
    )
    .parse(span)
}
