use crate::{ast::Expression, error::Error};
use std::collections::HashMap;

/// The arguments that may be passed to a spell for its instanciation.
#[derive(Debug, Default)]
pub struct SpellArguments {
    named: HashMap<String, Expression>,
    trailing: Vec<Expression>,
}

impl SpellArguments {
    pub fn new(arguments: impl IntoIterator<Item = (Option<String>, Expression)>) -> Self {
        let arguments = arguments.into_iter();
        let mut this = Self::empty_with_capacity({
            let (min, maybe_max) = arguments.size_hint();
            maybe_max.unwrap_or(min)
        });

        arguments.for_each(|(key, expr)| match key {
            Some(key) => _ = this.named.entry(key).or_insert(expr),
            None => this.trailing.push(expr),
        });

        this
    }

    fn empty_with_capacity(capacity: usize) -> Self {
        Self {
            named: HashMap::with_capacity(capacity),
            trailing: Vec::with_capacity(capacity),
        }
    }

    /// Removes a named argument of a [SpellArguments]. On success returns
    /// [Option<Argument>::Some], otherwise [None].
    pub fn remove(&mut self, name: impl AsRef<str>) -> Option<Expression> {
        self.named.remove(name.as_ref())
    }

    /// Like [Self::remove], but we make sure that only the passed names are present.
    pub fn remove_exact<const N: usize>(
        &mut self,
        names: [&str; N],
    ) -> Result<[Expression; N], Error> {
        if let Some(missing) = names.iter().find(|&&name| !self.named.contains_key(name)) {
            Err(Error::Undefined(Into::into(format!(
                "mendatory spell argument '{missing}'"
            ))))
        } else if self.named.len() > N {
            Err(Error::Unexpected(Into::into(
                "too many named arguments in in spell arguments",
            )))
        } else {
            Ok(names.map(|name| self.remove(name).unwrap()))
        }
    }

    /// Pops an unamed argument from the argument list.
    pub fn pop(&mut self) -> Option<Expression> {
        self.trailing.pop()
    }

    /// Iterate over the name of the named arguments.
    pub fn iter_named_keys(&self) -> impl Iterator<Item = &str> {
        self.named.keys().map(String::as_str)
    }

    /// Tells wether we have named arguments or not.
    pub fn has_named(&self) -> bool {
        self.named.is_empty()
    }

    /// Tells wether we have unamed arguments or not.
    pub fn has_unamed(&self) -> bool {
        self.trailing.is_empty()
    }

    /// Tells wether the argument list is empty or not. Note that at the end of the
    /// [crate::spell::Spell] creation, the argument list must be empty!
    pub fn is_empty(&self) -> bool {
        !self.has_named() && !self.has_unamed()
    }
}

/// Iterate over all the arguments. You are guarentied to have first the named arguments - in an
/// unspecified order - then the unamed arguments in order.
pub struct IntoIter {
    named: <HashMap<String, Expression> as IntoIterator>::IntoIter,
    trailing: <Vec<Expression> as IntoIterator>::IntoIter,
}

impl IntoIterator for SpellArguments {
    type Item = <Self::IntoIter as Iterator>::Item;
    type IntoIter = IntoIter;

    /// Iterate over all the arguments. You are guarentied to have first the named arguments - in
    /// an unspecified order - then the unamed arguments in order.
    fn into_iter(self) -> Self::IntoIter {
        IntoIter {
            named: self.named.into_iter(),
            trailing: self.trailing.into_iter(),
        }
    }
}

impl Iterator for IntoIter {
    type Item = (Option<String>, Expression);

    fn next(&mut self) -> Option<Self::Item> {
        self.named
            .next()
            .map(|(key, constant)| (Some(key), constant))
            .or_else(|| self.trailing.next().map(|constant| (None, constant)))
    }
}
