use crate::{
    error::Error,
    spell::{BuildableSpell, Spell, SpellArguments},
};
use std::collections::HashMap;

/// The builder function, call it with some arguments to try to build the spell
pub type SpellBuilderFunction = fn(SpellArguments) -> Result<Box<dyn Spell>, Error>;

/// The spell factory, you have to register the spell before trying to create them.
#[derive(Debug)]
pub struct SpellFactory {
    /// The dispatch map, to get a builder by an associated spell name.
    dispatch: HashMap<&'static str, SpellBuilderFunction>,
}

impl Default for SpellFactory {
    fn default() -> Self {
        use crate::spell::std as spell;
        Self::new_without_std().register::<spell::Log>()
    }
}

impl SpellFactory {
    /// Create a new spell factory with standard spells.
    pub fn new_without_std() -> Self {
        Self {
            dispatch: Default::default(),
        }
    }

    /// Get the list of all registered spells.
    pub fn get_registered_spells(&self) -> impl Iterator<Item = &'static str> {
        self.dispatch.keys().copied()
    }

    /// Register a spell into the factory. In case of re-registering returns an [Err].
    pub fn register<Sp: BuildableSpell + 'static>(mut self) -> Self {
        Sp::name_list().for_each(|name| _ = self.dispatch.entry(name).or_insert(Sp::create));
        self
    }

    /// Get the builder of a [Spell] by its name. In case of undefined [Spell], returns an [Err].
    pub fn get_builder(&self, name: impl AsRef<str>) -> Result<SpellBuilderFunction, Error> {
        self.dispatch
            .get(name.as_ref())
            .copied()
            .ok_or(Error::Undefined(name.as_ref().to_string().into()))
    }
}
