use crate::prelude::*;

#[derive(Debug)]
pub struct Log {
    if_predicate: Option<GrimoireExpression>,
    level: GrimoireExpression,
    messages: Vec<GrimoireExpression>,
}

impl GrimoireSpell for Log {
    fn cast(&self, state: GrimoireState) -> Result<GrimoireState, GrimoireError> {
        if let Some(pred) = &self.if_predicate {
            if !state
                .evaluate_as(pred, GrimoireType::Boolean)?
                .try_into_const()?
                .unwrap_boolean()
            {
                return Ok(state);
            }
        }

        let (heading, complementary) = match self.messages.split_first() {
            Some((heading, complementary)) => (heading, complementary),
            None => return Ok(state),
        };

        let level = state
            .evaluate_as(&self.level, GrimoireType::String)?
            .try_into_const()?
            .to_string();

        let logger = |ident: bool, line: &GrimoireExpression| -> Result<(), GrimoireError> {
            let ident = ident.then_some("\t").unwrap_or_default();
            let line = state
                .evaluate_as(line, GrimoireType::String)?
                .try_into_const()?
                .to_string();

            match level.as_str() {
                "error" => log::error!(target: "grimoire", "{ident}{line}"),
                _ => log::info!(target: "grimoire", "{ident}{line}"),
            }

            Ok(())
        };

        logger(false, heading)?;
        complementary
            .iter()
            .try_for_each(|complementary| logger(true, complementary))?;

        Ok(state)
    }
}

impl TryFrom<spell::Arguments> for Log {
    type Error = GrimoireError;

    fn try_from(mut args: spell::Arguments) -> Result<Self, Self::Error> {
        let if_predicate = args.remove("if");
        let [level] = args.remove_exact(["level"])?;
        let messages = args
            .into_iter()
            .filter_map(|(key, msg)| key.is_none().then_some(msg))
            .collect();

        Ok(Self {
            if_predicate,
            level,
            messages,
        })
    }
}

impl GrimoireBuildableSpell for Log {
    fn name() -> &'static str {
        "log"
    }
}
