use crate::{error::Error, spell::SpellArguments, state::State};

/// A spell, can be casted on a state to modify it.
pub trait Spell {
    /// Cast the spell on a [State]. Returns a new [State].
    fn cast(&self, state: State) -> Result<State, Error>;
}

/// A [Spell] that we can create. Is used to be able to register it into a
/// [crate::spell::SpellFactory].
pub trait BuildableSpell
where
    Self: Spell + TryFrom<SpellArguments, Error = Error> + 'static,
{
    /// The name of the spell.
    fn name() -> &'static str;

    /// Aliases for this spell.
    fn aliases() -> &'static [&'static str] {
        &[]
    }

    /// The list of names: the name and all the aliases.
    fn name_list() -> impl Iterator<Item = &'static str> {
        [Self::name()]
            .into_iter()
            .chain(Self::aliases().iter().copied())
    }

    /// Create a spell from a set of arguments. May fail.
    fn create(arguments: SpellArguments) -> Result<Box<dyn Spell>, Error> {
        Self::try_from(arguments)
            .map(|this| Box::new(this) as Box<dyn Spell>)
            .map_err(|err| Error::BuildSpell(Self::name(), Box::new(err)))
    }
}
