use crate::{
    ast::{Const, Expression, VarOrConst, operator},
    error::Error,
    types::Type,
    vn::{Actor, ConstOrVnType, Printer},
};
use std::{
    collections::{HashMap, HashSet},
    rc::Rc,
};

#[derive(Default)]
pub struct State {
    variables: HashMap<String, Const>,
    identifiers: HashSet<String>,
    actors: HashMap<String, Rc<dyn Actor>>,
    printers: HashMap<String, Rc<dyn Printer>>,
}

impl State {
    pub fn resolve_variable(&self, var: impl AsRef<str>) -> Result<&Const, Error> {
        self.variables
            .get(var.as_ref())
            .ok_or_else(|| Error::Undefined(var.as_ref().to_string().into()))
    }

    fn evaluate_numeric_binop(
        &self,
        left: impl AsRef<Expression>,
        op: operator::Numeric,
        right: impl AsRef<Expression>,
    ) -> Result<Const, Error> {
        let left = self.evaluate_as(left, Type::Number)?.try_into_const()?;
        let right = self.evaluate_as(right, Type::Number)?.try_into_const()?;

        match matches!(left, Const::Flt(_)) || matches!(right, Const::Flt(_)) {
            // Use float operations
            true => {
                let left = left.unwrap_float();
                let right = right.unwrap_float();
                match op {
                    operator::Numeric::Add => Ok(Const::Flt(left + right)),
                    operator::Numeric::Sub => Ok(Const::Flt(left - right)),
                    operator::Numeric::Mul => Ok(Const::Flt(left * right)),
                    operator::Numeric::CompGE => Ok(Const::Bool(left >= right)),
                    operator::Numeric::CompGT => Ok(Const::Bool(left > right)),
                    operator::Numeric::CompLT => Ok(Const::Bool(left < right)),
                    operator::Numeric::CompLE => Ok(Const::Bool(left <= right)),

                    operator::Numeric::Div if right.abs() >= f32::EPSILON => {
                        Ok(Const::Flt(left / right))
                    }
                    operator::Numeric::Mod if right.abs() >= f32::EPSILON => {
                        Ok(Const::Flt(left % right))
                    }

                    operator::Numeric::Div => Ok(Const::Flt(f32::NAN)),
                    operator::Numeric::Mod => Ok(Const::Flt(f32::NAN)),
                }
            }

            // Use integer operations
            false => {
                let left = left.unwrap_integer();
                let right = right.unwrap_integer();
                match op {
                    operator::Numeric::Add => Ok(Const::Int(left + right)),
                    operator::Numeric::Sub => Ok(Const::Int(left - right)),
                    operator::Numeric::Mul => Ok(Const::Int(left * right)),
                    operator::Numeric::CompGE => Ok(Const::Bool(left >= right)),
                    operator::Numeric::CompGT => Ok(Const::Bool(left > right)),
                    operator::Numeric::CompLT => Ok(Const::Bool(left < right)),
                    operator::Numeric::CompLE => Ok(Const::Bool(left <= right)),

                    operator::Numeric::Div if right != 0 => Ok(Const::Int(left / right)),
                    operator::Numeric::Mod if right != 0 => Ok(Const::Int(left % right)),

                    operator::Numeric::Div => Ok(Const::Flt(f32::NAN)),
                    operator::Numeric::Mod => Ok(Const::Flt(f32::NAN)),
                }
            }
        }
    }

    fn evaluate_logic_binop(
        &self,
        left: impl AsRef<Expression>,
        op: operator::Logic,
        right: impl AsRef<Expression>,
    ) -> Result<Const, Error> {
        // Here we evaluate eagerly to catch errors!
        let left = self
            .evaluate_as(left, Type::Boolean)?
            .try_into_const()?
            .unwrap_boolean();
        let right = self
            .evaluate_as(right, Type::Boolean)?
            .try_into_const()?
            .unwrap_boolean();
        Ok(Const::Bool(match op {
            operator::Logic::Or => left || right,
            operator::Logic::And => left && right,
            operator::Logic::Xor => left ^ right,
        }))
    }

    fn evaluate_equal_binop(
        &self,
        left: impl AsRef<Expression>,
        op: operator::Equal,
        right: impl AsRef<Expression>,
    ) -> Result<Const, Error> {
        let left = self.evaluate(left)?.try_into_const()?;
        let right = self.evaluate(right)?.try_into_const()?;
        Ok(Const::Bool(match (left, right) {
            // Easy!
            (Const::Int(x), Const::Int(y)) => op.apply(x == y),
            (Const::Flt(x), Const::Flt(y)) => op.apply(x == y),
            (Const::Bool(x), Const::Bool(y)) => op.apply(x == y),

            (Const::Str(x), Const::Ident(y))
            | (Const::Ident(x), Const::Str(y))
            | (Const::Ident(x), Const::Ident(y))
            | (Const::Str(x), Const::Str(y)) => op.apply(x == y),

            // Cast, handle permutations and cast priority.
            (Const::Int(x), Const::Flt(y)) => op.apply(x == y as i32),
            (Const::Flt(x), Const::Int(y)) => op.apply(x == y as f32),

            // Cast with permutation. Here we cast all into a boolean first!
            (x @ Const::Int(_), Const::Bool(y)) | (Const::Bool(y), x @ Const::Int(_)) => {
                op.apply(x.unwrap_boolean() == y)
            }

            (x @ Const::Flt(_), Const::Bool(y)) | (Const::Bool(y), x @ Const::Flt(_)) => {
                op.apply(x.unwrap_boolean() == y)
            }

            // Can be inter-casted, they are not equal!
            _ => op.check_not_equals(),
        }))
    }

    pub fn evaluate(&self, expr: impl AsRef<Expression>) -> Result<ConstOrVnType, Error> {
        match expr.as_ref() {
            // Leaf, simple.
            Expression::Leaf(_, VarOrConst::Const(constant)) => Ok(constant.clone().into()),
            Expression::Leaf(_, VarOrConst::Var(var)) => {
                self.resolve_variable(var).cloned().map(Into::into)
            }

            // Unary, not complicated.
            Expression::Unary(_, op, inner) => match op {
                operator::Unary::Not => Ok(Const::Bool(
                    !self
                        .evaluate_as(inner, Type::Boolean)?
                        .try_into_const()?
                        .unwrap_boolean(),
                )),
                operator::Unary::Neg => {
                    match self.evaluate_as(inner, Type::Number)?.try_into_const()? {
                        Const::Int(x) => Ok(Const::Int(-x)),
                        Const::Flt(x) => Ok(Const::Flt(-x)),
                        _ => unreachable!(),
                    }
                }
            }
            .map(Into::into),

            // Binary, just long to write...
            Expression::Binary(_, left, op, right) => match op {
                operator::Binary::Numeric(op) => self.evaluate_numeric_binop(left, *op, right),
                operator::Binary::Logic(op) => self.evaluate_logic_binop(left, *op, right),
                operator::Binary::Equal(op) => self.evaluate_equal_binop(left, *op, right),
            }
            .map(Into::into),
        }
    }

    fn check_identifier(&self, ident: impl Into<String>) -> Result<Const, Error> {
        let x = ident.into();
        match self.identifiers.contains(&x) {
            true => Ok(Const::Ident(x)),
            false => Err(Error::Undefined(x.into())),
        }
    }

    fn find_actor(&self, actor: impl Into<String>) -> Result<Rc<dyn Actor>, Error> {
        let actor = actor.into();
        self.actors
            .get(&actor)
            .cloned()
            .ok_or(Error::Undefined(actor.into()))
    }

    fn find_printer(&self, printer: impl Into<String>) -> Result<Rc<dyn Printer>, Error> {
        let printer = printer.into();
        self.printers
            .get(&printer)
            .cloned()
            .ok_or(Error::Undefined(printer.into()))
    }

    pub fn evaluate_as(
        &self,
        expr: impl AsRef<Expression>,
        ty: Type,
    ) -> Result<ConstOrVnType, Error> {
        let res = self.evaluate(expr)?;

        match ty {
            Type::Identifier => self
                .check_identifier(res.try_into_const()?.unwrap_identifier())
                .map(ConstOrVnType::Const),

            Type::Actor => match res.is_actor() {
                true => Ok(res),
                false => self
                    .find_actor(res.try_into_const()?.unwrap_identifier())
                    .map(ConstOrVnType::Actor),
            },

            Type::Printer => match res.is_printer() {
                true => Ok(res),
                false => self
                    .find_printer(res.try_into_const()?.unwrap_identifier())
                    .map(ConstOrVnType::Printer),
            },

            ty => res
                .try_into_const()?
                .cast_into(ty)
                .map(ConstOrVnType::Const),
        }
    }
}
